@@ -87,7 +87,7 @@ case "${LSBNAME}" in
     ;&
   focal)
     ;&
-  impish|jammy)
+  impish|jammy|kinetic)
     # mariadb-plugin-rocksdb s390x not supported by us (yet)
     # ubuntu doesn't support mips64el yet, so keep this just
     # in case something changes.
