@@ -15,7 +15,7 @@
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1335  USA */
 
 #define MYSQL_LEX 1
-#include <my_global.h>                          /* NO_EMBEDDED_ACCESS_CHECKS */
+#include "mariadb.h"                          /* NO_EMBEDDED_ACCESS_CHECKS */
 #include "sql_priv.h"
 #include "unireg.h"
 #include "sql_parse.h"                          // parse_sql
@@ -25,19 +25,34 @@
                                            // date_add_interval,
                                            // calc_time_diff
 #include "tztime.h"     // my_tz_find, my_tz_OFFSET0, struct Time_zone
-#include "sql_acl.h"    // EVENT_ACL, SUPER_ACL
 #include "sp.h"         // load_charset, load_collation
 #include "events.h"
 #include "event_data_objects.h"
 #include "event_db_repository.h"
 #include "sp_head.h"
 #include "sql_show.h"                // append_definer, append_identifier
-
+#include "mysql/psi/mysql_sp.h"
+#include "wsrep_mysqld.h"
+#ifdef WITH_WSREP
+#include "wsrep_trans_observer.h"
+#endif /* WITH_WSREP */
 /**
   @addtogroup Event_Scheduler
   @{
 */
 
+#ifdef HAVE_PSI_INTERFACE
+void init_scheduler_psi_keys()
+{
+  const char *category= "scheduler";
+
+  PSI_server->register_statement(category, & Event_queue_element_for_exec::psi_info, 1);
+}
+
+PSI_statement_info Event_queue_element_for_exec::psi_info=
+{ 0, "event", 0};
+#endif
+
 /*************************************************************************/
 
 /**
@@ -171,13 +186,16 @@ Event_creation_ctx::load_from_db(THD *thd,
 */
 
 bool
-Event_queue_element_for_exec::init(LEX_STRING db, LEX_STRING n)
+Event_queue_element_for_exec::init(const LEX_CSTRING &db, const LEX_CSTRING &n)
 {
-  if (!(dbname.str= my_strndup(db.str, dbname.length= db.length, MYF(MY_WME))))
+  if (!(dbname.str= my_strndup(key_memory_Event_queue_element_for_exec_names,
+                               db.str, dbname.length= db.length, MYF(MY_WME))))
     return TRUE;
-  if (!(name.str= my_strndup(n.str, name.length= n.length, MYF(MY_WME))))
+  if (!(name.str= my_strndup(key_memory_Event_queue_element_for_exec_names,
+                             n.str, name.length= n.length, MYF(MY_WME))))
   {
-    my_free(dbname.str);
+    my_free(const_cast<char*>(dbname.str));
+    dbname.str= NULL;
     return TRUE;
   }
   return FALSE;
@@ -193,8 +211,8 @@ Event_queue_element_for_exec::init(LEX_STRING db, LEX_STRING n)
 
 Event_queue_element_for_exec::~Event_queue_element_for_exec()
 {
-  my_free(dbname.str);
-  my_free(name.str);
+  my_free(const_cast<char*>(dbname.str));
+  my_free(const_cast<char*>(name.str));
 }
 
 
@@ -209,7 +227,7 @@ Event_basic::Event_basic()
 {
   DBUG_ENTER("Event_basic::Event_basic");
   /* init memory root */
-  init_sql_alloc(&mem_root, 256, 512, MYF(0));
+  init_sql_alloc(key_memory_event_basic_root, &mem_root, 256, 512, MYF(0));
   dbname.str= name.str= NULL;
   dbname.length= name.length= 0;
   time_zone= NULL;
@@ -233,7 +251,7 @@ Event_basic::~Event_basic()
 
 
 /*
-  Short function to load a char column into a LEX_STRING
+  Short function to load a char column into a LEX_CSTRING
 
   SYNOPSIS
     Event_basic::load_string_field()
@@ -249,7 +267,7 @@ Event_basic::load_string_fields(Field **fields, ...)
   bool ret= FALSE;
   va_list args;
   enum enum_events_table_field field_name;
-  LEX_STRING *field_value;
+  LEX_CSTRING *field_value;
 
   DBUG_ENTER("Event_basic::load_string_fields");
 
@@ -257,7 +275,7 @@ Event_basic::load_string_fields(Field **fields, ...)
   field_name= (enum enum_events_table_field) va_arg(args, int);
   while (field_name < ET_FIELD_COUNT)
   {
-    field_value= va_arg(args, LEX_STRING *);
+    field_value= va_arg(args, LEX_CSTRING *);
     if ((field_value->str= get_field(&mem_root, fields[field_name])) == NullS)
     {
       ret= TRUE;
@@ -274,9 +292,9 @@ Event_basic::load_string_fields(Field **fields, ...)
 
 
 bool
-Event_basic::load_time_zone(THD *thd, const LEX_STRING tz_name)
+Event_basic::load_time_zone(THD *thd, const LEX_CSTRING *tz_name)
 {
-  String str(tz_name.str, &my_charset_latin1);
+  String str(tz_name->str, &my_charset_latin1);
   time_zone= my_tz_find(thd, &str);
 
   return (time_zone == NULL);
@@ -391,9 +409,9 @@ Event_timed::init()
 bool
 Event_job_data::load_from_row(THD *thd, TABLE *table)
 {
-  char *ptr;
+  const char *ptr;
   size_t len;
-  LEX_STRING tz_name;
+  LEX_CSTRING tz_name;
 
   DBUG_ENTER("Event_job_data::load_from_row");
 
@@ -412,7 +430,7 @@ Event_job_data::load_from_row(THD *thd, TABLE *table)
                          ET_FIELD_COUNT))
     DBUG_RETURN(TRUE);
 
-  if (load_time_zone(thd, tz_name))
+  if (load_time_zone(thd, &tz_name))
     DBUG_RETURN(TRUE);
 
   Event_creation_ctx::load_from_db(thd, &mem_root, dbname.str, name.str, table,
@@ -431,7 +449,7 @@ Event_job_data::load_from_row(THD *thd, TABLE *table)
   definer_host.str= strmake_root(&mem_root, ptr + 1, len);
   definer_host.length= len;
 
-  sql_mode= (ulong) table->field[ET_FIELD_SQL_MODE]->val_int();
+  sql_mode= (sql_mode_t) table->field[ET_FIELD_SQL_MODE]->val_int();
 
   DBUG_RETURN(FALSE);
 }
@@ -452,9 +470,9 @@ Event_job_data::load_from_row(THD *thd, TABLE *table)
 bool
 Event_queue_element::load_from_row(THD *thd, TABLE *table)
 {
-  char *ptr;
+  const char *ptr;
   MYSQL_TIME time;
-  LEX_STRING tz_name;
+  LEX_CSTRING tz_name;
 
   DBUG_ENTER("Event_queue_element::load_from_row");
 
@@ -472,21 +490,31 @@ Event_queue_element::load_from_row(THD *thd, TABLE *table)
                          ET_FIELD_COUNT))
     DBUG_RETURN(TRUE);
 
-  if (load_time_zone(thd, tz_name))
+  if (load_time_zone(thd, &tz_name))
     DBUG_RETURN(TRUE);
 
   starts_null= table->field[ET_FIELD_STARTS]->is_null();
   uint not_used;
   if (!starts_null)
   {
-    table->field[ET_FIELD_STARTS]->get_date(&time, TIME_NO_ZERO_DATE);
+    /*
+      The expected data type for these columns in mysql.events:
+        starts, ends, execute_at, last_executed
+      is DATETIME. No nanosecond truncation should normally be needed,
+      unless the DBA changes them, e.g. to VARCHAR, DECIMAL, etc.
+      For this unexpected case let's use the default round mode,
+      according to the current session settings.
+    */
+    table->field[ET_FIELD_STARTS]->get_date(&time, TIME_NO_ZERO_DATE |
+                                                   thd->temporal_round_mode());
     starts= my_tz_OFFSET0->TIME_to_gmt_sec(&time,&not_used);
   }
 
   ends_null= table->field[ET_FIELD_ENDS]->is_null();
   if (!ends_null)
   {
-    table->field[ET_FIELD_ENDS]->get_date(&time, TIME_NO_ZERO_DATE);
+    table->field[ET_FIELD_ENDS]->get_date(&time, TIME_NO_ZERO_DATE |
+                                                 thd->temporal_round_mode());
     ends= my_tz_OFFSET0->TIME_to_gmt_sec(&time,&not_used);
   }
 
@@ -502,8 +530,8 @@ Event_queue_element::load_from_row(THD *thd, TABLE *table)
   DBUG_ASSERT(!(starts_null && ends_null && !expression && execute_at_null));
   if (!expression && !execute_at_null)
   {
-    if (table->field[ET_FIELD_EXECUTE_AT]->get_date(&time,
-                                                    TIME_NO_ZERO_DATE))
+    if (table->field[ET_FIELD_EXECUTE_AT]->get_date(&time, TIME_NO_ZERO_DATE |
+                                                    thd->temporal_round_mode()))
       DBUG_RETURN(TRUE);
     execute_at= my_tz_OFFSET0->TIME_to_gmt_sec(&time,&not_used);
   }
@@ -519,7 +547,7 @@ Event_queue_element::load_from_row(THD *thd, TABLE *table)
     int i;
     char buff[MAX_FIELD_WIDTH];
     String str(buff, sizeof(buff), &my_charset_bin);
-    LEX_STRING tmp;
+    LEX_CSTRING tmp;
 
     table->field[ET_FIELD_TRANSIENT_INTERVAL]->val_str(&str);
     if (!(tmp.length= str.length()))
@@ -535,8 +563,8 @@ Event_queue_element::load_from_row(THD *thd, TABLE *table)
 
   if (!table->field[ET_FIELD_LAST_EXECUTED]->is_null())
   {
-    table->field[ET_FIELD_LAST_EXECUTED]->get_date(&time,
-                                                   TIME_NO_ZERO_DATE);
+    table->field[ET_FIELD_LAST_EXECUTED]->get_date(&time, TIME_NO_ZERO_DATE |
+                                                   thd->temporal_round_mode());
     last_executed= my_tz_OFFSET0->TIME_to_gmt_sec(&time,&not_used);
   }
 
@@ -590,7 +618,7 @@ Event_queue_element::load_from_row(THD *thd, TABLE *table)
 bool
 Event_timed::load_from_row(THD *thd, TABLE *table)
 {
-  char *ptr;
+  const char *ptr;
   size_t len;
 
   DBUG_ENTER("Event_timed::load_from_row");
@@ -637,7 +665,7 @@ Event_timed::load_from_row(THD *thd, TABLE *table)
   else
     comment.length= 0;
 
-  sql_mode= (ulong) table->field[ET_FIELD_SQL_MODE]->val_int();
+  sql_mode= (sql_mode_t) table->field[ET_FIELD_SQL_MODE]->val_int();
 
   DBUG_RETURN(FALSE);
 }
@@ -654,7 +682,7 @@ my_time_t
 add_interval(MYSQL_TIME *ltime, const Time_zone *time_zone,
              interval_type scale, INTERVAL interval)
 {
-  if (date_add_interval(ltime, scale, interval))
+  if (date_add_interval(current_thd, ltime, scale, interval))
     return 0;
 
   uint not_used;
@@ -758,8 +786,8 @@ bool get_next_time(const Time_zone *time_zone, my_time_t *next,
 
   if (seconds)
   {
-    longlong seconds_diff;
-    long microsec_diff;
+    ulonglong seconds_diff;
+    ulong microsec_diff;
     bool negative= calc_time_diff(&local_now, &local_start, 1,
                                   &seconds_diff, &microsec_diff);
     if (!negative)
@@ -1196,14 +1224,14 @@ Event_timed::get_create_event(THD *thd, String *buf)
   buf->append(STRING_WITH_LEN("CREATE "));
   append_definer(thd, buf, &definer_user, &definer_host);
   buf->append(STRING_WITH_LEN("EVENT "));
-  append_identifier(thd, buf, name.str, name.length);
+  append_identifier(thd, buf, &name);
 
   if (expression)
   {
     buf->append(STRING_WITH_LEN(" ON SCHEDULE EVERY "));
     buf->append(expr_buf);
     buf->append(' ');
-    LEX_STRING *ival= &interval_type_to_name[interval];
+    LEX_CSTRING *ival= &interval_type_to_name[interval];
     buf->append(ival->str, ival->length);
 
     if (!starts_null)
@@ -1236,7 +1264,7 @@ Event_timed::get_create_event(THD *thd, String *buf)
     append_unescaped(buf, comment.str, comment.length);
   }
   buf->append(STRING_WITH_LEN(" DO "));
-  buf->append(body.str, body.length);
+  buf->append(&body);
 
   DBUG_RETURN(0);
 }
@@ -1249,7 +1277,7 @@ Event_timed::get_create_event(THD *thd, String *buf)
 bool
 Event_job_data::construct_sp_sql(THD *thd, String *sp_sql)
 {
-  LEX_STRING buffer;
+  LEX_CSTRING buffer;
   const uint STATIC_SQL_LENGTH= 44;
 
   DBUG_ENTER("Event_job_data::construct_sp_sql");
@@ -1266,15 +1294,15 @@ Event_job_data::construct_sp_sql(THD *thd, String *sp_sql)
   sp_sql->length(0);
 
 
-  sp_sql->append(C_STRING_WITH_LEN("CREATE "));
-  sp_sql->append(C_STRING_WITH_LEN("PROCEDURE "));
+  sp_sql->append(STRING_WITH_LEN("CREATE "));
+  sp_sql->append(STRING_WITH_LEN("PROCEDURE "));
   /*
     Let's use the same name as the event name to perhaps produce a
     better error message in case it is a part of some parse error.
     We're using append_identifier here to successfully parse
     events with reserved names.
   */
-  append_identifier(thd, sp_sql, name.str, name.length);
+  append_identifier(thd, sp_sql, &name);
 
   /*
     The default SQL security of a stored procedure is DEFINER. We
@@ -1282,9 +1310,13 @@ Event_job_data::construct_sp_sql(THD *thd, String *sp_sql)
     let's execute the procedure with the invoker rights to save on
     resets of security contexts.
   */
-  sp_sql->append(C_STRING_WITH_LEN("() SQL SECURITY INVOKER "));
+  sp_sql->append(STRING_WITH_LEN("() SQL SECURITY INVOKER "));
 
-  sp_sql->append(body.str, body.length);
+  if (thd->variables.sql_mode & MODE_ORACLE)
+    sp_sql->append(STRING_WITH_LEN(" AS BEGIN "));
+  sp_sql->append(&body);
+  if (thd->variables.sql_mode & MODE_ORACLE)
+    sp_sql->append(STRING_WITH_LEN("; END"));
 
   DBUG_RETURN(thd->is_fatal_error);
 }
@@ -1298,7 +1330,7 @@ Event_job_data::construct_sp_sql(THD *thd, String *sp_sql)
 bool
 Event_job_data::construct_drop_event_sql(THD *thd, String *sp_sql)
 {
-  LEX_STRING buffer;
+  LEX_CSTRING buffer;
   const uint STATIC_SQL_LENGTH= 14;
 
   DBUG_ENTER("Event_job_data::construct_drop_event_sql");
@@ -1310,10 +1342,10 @@ Event_job_data::construct_drop_event_sql(THD *thd, String *sp_sql)
   sp_sql->set(buffer.str, buffer.length, system_charset_info);
   sp_sql->length(0);
 
-  sp_sql->append(C_STRING_WITH_LEN("DROP EVENT "));
-  append_identifier(thd, sp_sql, dbname.str, dbname.length);
+  sp_sql->append(STRING_WITH_LEN("DROP EVENT "));
+  append_identifier(thd, sp_sql, &dbname);
   sp_sql->append('.');
-  append_identifier(thd, sp_sql, name.str, name.length);
+  append_identifier(thd, sp_sql, &name);
 
   DBUG_RETURN(thd->is_fatal_error);
 }
@@ -1339,6 +1371,10 @@ Event_job_data::execute(THD *thd, bool drop)
 
   thd->reset_for_next_command();
 
+#ifdef WITH_WSREP
+  wsrep_open(thd);
+  wsrep_before_command(thd);
+#endif /* WITH_WSREP */
   /*
     MySQL parser currently assumes that current database is either
     present in THD or all names in all statements are fully specified.
@@ -1355,7 +1391,7 @@ Event_job_data::execute(THD *thd, bool drop)
     mysql_change_db will be invoked anyway later, to activate the
     procedure database before it's executed.
   */
-  thd->set_db(dbname.str, dbname.length);
+  thd->set_db(&dbname);
 
   lex_start(thd);
 
@@ -1387,9 +1423,6 @@ Event_job_data::execute(THD *thd, bool drop)
     goto end;
   }
 
-  if (construct_sp_sql(thd, &sp_sql))
-    goto end;
-
   /*
     Set up global thread attributes to reflect the properties of
     this Event. We can simply reset these instead of usual
@@ -1401,19 +1434,30 @@ Event_job_data::execute(THD *thd, bool drop)
   thd->variables.sql_mode= sql_mode;
   thd->variables.time_zone= time_zone;
 
+  if (construct_sp_sql(thd, &sp_sql))
+    goto end;
+
   thd->set_query(sp_sql.c_ptr_safe(), sp_sql.length());
 
   {
     Parser_state parser_state;
+    sql_digest_state *parent_digest= thd->m_digest;
+    PSI_statement_locker *parent_locker= thd->m_statement_psi;
+    bool res;
+
     if (parser_state.init(thd, thd->query(), thd->query_length()))
       goto end;
 
-    if (parse_sql(thd, & parser_state, creation_ctx))
+    thd->m_digest= NULL;
+    thd->m_statement_psi= NULL;
+    res= parse_sql(thd, & parser_state, creation_ctx);
+    thd->m_digest= parent_digest;
+    thd->m_statement_psi= parent_locker;
+
+    if (res)
     {
-      sql_print_error("Event Scheduler: "
-                      "%serror during compilation of %s.%s",
-                      thd->is_fatal_error ? "fatal " : "",
-                      (const char *) dbname.str, (const char *) name.str);
+      sql_print_error("Event Scheduler: %serror during compilation of %s.%s",
+                      thd->is_fatal_error ? "fatal " : "", dbname.str, name.str);
       goto end;
     }
   }
@@ -1426,10 +1470,19 @@ Event_job_data::execute(THD *thd, bool drop)
     sphead->m_flags|= sp_head::LOG_SLOW_STATEMENTS;
     sphead->m_flags|= sp_head::LOG_GENERAL_LOG;
 
-    sphead->set_info(0, 0, &thd->lex->sp_chistics, sql_mode);
+    /*
+      construct_sp_sql() + parse_sql() set suid to SP_IS_NOT_SUID,
+      because we have the security context already set to the event
+      definer here. See more comments in construct_sp_sql().
+    */
+    DBUG_ASSERT(sphead->suid() == SP_IS_NOT_SUID);
+    sphead->m_sql_mode= sql_mode;
     sphead->set_creation_ctx(creation_ctx);
     sphead->optimize();
 
+    sphead->m_sp_share= MYSQL_GET_SP_SHARE(SP_TYPE_EVENT,
+                                           dbname.str, static_cast<uint>(dbname.length),
+                                           name.str, static_cast<uint>(name.length));
     ret= sphead->execute_procedure(thd, &empty_item_list);
     /*
       There is no pre-locking and therefore there should be no
@@ -1438,7 +1491,7 @@ Event_job_data::execute(THD *thd, bool drop)
   }
 
 end:
-  if (drop && !thd->is_fatal_error)
+  if (drop && likely(!thd->is_fatal_error))
   {
     /*
       We must do it here since here we're under the right authentication
@@ -1454,8 +1507,6 @@ end:
       ret= 1;
     else
     {
-      ulong saved_master_access;
-
       thd->set_query(sp_sql.c_ptr_safe(), sp_sql.length());
 
       /*
@@ -1467,8 +1518,8 @@ end:
         Temporarily reset it to read-write.
       */
 
-      saved_master_access= thd->security_ctx->master_access;
-      thd->security_ctx->master_access |= SUPER_ACL;
+      privilege_t saved_master_access(thd->security_ctx->master_access);
+      thd->security_ctx->master_access |= PRIV_IGNORE_READ_ONLY;
       bool save_tx_read_only= thd->tx_read_only;
       thd->tx_read_only= false;
 
@@ -1491,7 +1542,7 @@ end:
       if (sql_command_set)
         thd->lex->sql_command = SQLCOM_DROP_EVENT;
 
-      ret= Events::drop_event(thd, dbname, name, FALSE);
+      ret= Events::drop_event(thd, &dbname, &name, FALSE);
 
       if (sql_command_set)
       {
@@ -1507,6 +1558,10 @@ end:
   if (save_sctx)
     event_sctx.restore_security_context(thd, save_sctx);
 #endif
+#ifdef WITH_WSREP
+  wsrep_after_command_ignore_result(thd);
+  wsrep_close(thd);
+#endif /* WITH_WSREP */
   thd->lex->unit.cleanup();
   thd->end_statement();
   thd->cleanup_after_query();
@@ -1533,9 +1588,9 @@ end:
 */
 
 bool
-event_basic_db_equal(LEX_STRING db, Event_basic *et)
+event_basic_db_equal(const LEX_CSTRING *db, Event_basic *et)
 {
-  return !sortcmp_lex_string(et->dbname, db, system_charset_info);
+  return !sortcmp_lex_string(&et->dbname, db, system_charset_info);
 }
 
 
@@ -1554,10 +1609,11 @@ event_basic_db_equal(LEX_STRING db, Event_basic *et)
 */
 
 bool
-event_basic_identifier_equal(LEX_STRING db, LEX_STRING name, Event_basic *b)
+event_basic_identifier_equal(const LEX_CSTRING *db, const LEX_CSTRING *name,
+                             Event_basic *b)
 {
-  return !sortcmp_lex_string(name, b->name, system_charset_info) &&
-         !sortcmp_lex_string(db, b->dbname, system_charset_info);
+  return !sortcmp_lex_string(name, &b->name, system_charset_info) &&
+         !sortcmp_lex_string(db, &b->dbname, system_charset_info);
 }
 
 /**
