@@ -15,12 +15,12 @@
 #include <stdio.h>
 #include <fcntl.h>
 #include <errno.h>
-#if defined(__WIN__)
+#if defined(_WIN32)
 #include <io.h>
 #include <winsock2.h>
 //#include <windows.h>
 #include <comdef.h>
-#else   // !__WIN__
+#else   // !_WIN32
 #include <sys/types.h>
 #include <sys/socket.h>
 #include <netinet/in.h>
@@ -28,7 +28,7 @@
 //#include <ctype.h>
 #include "osutil.h"
 #define _O_RDONLY O_RDONLY
-#endif  // !__WIN__
+#endif  // !_WIN32
 #include "resource.h"                        // for IDS_COLUMNS
 
 #define INCLUDE_TDBXML
@@ -53,11 +53,11 @@
 
 extern "C" char version[];
 
-#if defined(__WIN__) && defined(DOMDOC_SUPPORT)
+#if defined(_WIN32) && defined(DOMDOC_SUPPORT)
 #define XMLSUP "MS-DOM"
-#else   // !__WIN__
+#else   // !_WIN32
 #define XMLSUP "libxml2"
-#endif  // !__WIN__
+#endif  // !_WIN32
 
 #define TYPE_UNKNOWN     12        /* Must be greater than other types */
 #define XLEN(M)  sizeof(M) - strlen(M) - 1	       /* To avoid overflow*/
@@ -148,14 +148,21 @@ PQRYRES XMLColumns(PGLOBAL g, char *db, char *tab, PTOS topt, bool info)
   /*  Open the input file.                                             */
   /*********************************************************************/
   if (!(fn = GetStringTableOption(g, topt, "Filename", NULL))) {
-    strcpy(g->Message, MSG(MISSING_FNAME));
-    return NULL;
-  } else {
-    lvl = GetIntegerTableOption(g, topt, "Level", GetDefaultDepth());
-		lvl = GetIntegerTableOption(g, topt, "Depth", lvl);
-		lvl = (lvl < 0) ? 0 : (lvl > 16) ? 16 : lvl;
+    if (topt->http) // REST table can have default filename
+      fn = GetStringTableOption(g, topt, "Subtype", NULL);
+    
+    if (!fn) {
+      strcpy(g->Message, MSG(MISSING_FNAME));
+      return NULL;
+    } else
+      topt->subtype = NULL;
+
   } // endif fn
 
+  lvl = GetIntegerTableOption(g, topt, "Level", GetDefaultDepth());
+  lvl = GetIntegerTableOption(g, topt, "Depth", lvl);
+  lvl = (lvl < 0) ? 0 : (lvl > 16) ? 16 : lvl;
+
   if (trace(1))
     htrc("File %s lvl=%d\n", topt->filename, lvl);
 
@@ -173,11 +180,11 @@ PQRYRES XMLColumns(PGLOBAL g, char *db, char *tab, PTOS topt, bool info)
 	tdp->Skip = GetBooleanTableOption(g, topt, "Skipnull", false);
 
   if (!(op = GetStringTableOption(g, topt, "Xmlsup", NULL)))
-#if defined(__WIN__)
+#if defined(_WIN32)
     tdp->Usedom = true;
-#else   // !__WIN__
+#else   // !_WIN32
     tdp->Usedom = false;
-#endif  // !__WIN__
+#endif  // !_WIN32
   else
     tdp->Usedom = (toupper(*op) == 'M' || toupper(*op) == 'D');
 
@@ -509,7 +516,7 @@ bool XMLDEF::DefineAM(PGLOBAL g, LPCSTR am, int poff)
       defcol = "TD";
       break;
     default:
-      sprintf(g->Message, MSG(INV_COL_TYPE), buf);
+      snprintf(g->Message, sizeof(g->Message), MSG(INV_COL_TYPE), buf);
       return true;
     } // endswitch typname
 
@@ -522,7 +529,7 @@ bool XMLDEF::DefineAM(PGLOBAL g, LPCSTR am, int poff)
   XmlDB = GetStringCatInfo(g, "XmlDB", NULL);
   Nslist = GetStringCatInfo(g, "Nslist", NULL);
   DefNs = GetStringCatInfo(g, "DefNs", NULL);
-  Limit = GetIntCatInfo("Limit", 10);
+  Limit = GetIntCatInfo("Limit", 50);
   Xpand = GetBoolCatInfo("Expand", false);
   Header = GetIntCatInfo("Header", 0);
   GetCharCatInfo("Xmlsup", "*", buf, sizeof(buf));
@@ -530,11 +537,11 @@ bool XMLDEF::DefineAM(PGLOBAL g, LPCSTR am, int poff)
   // Note that if no support is specified, the default is MS-DOM
   // on Windows and libxml2 otherwise
   if (*buf == '*')
-#if defined(__WIN__)
+#if defined(_WIN32)
     Usedom = true;
-#else   // !__WIN__
+#else   // !_WIN32
     Usedom = false;
-#endif  // !__WIN__
+#endif  // !_WIN32
   else
     Usedom = (toupper(*buf) == 'M' || toupper(*buf) == 'D');
 
@@ -774,7 +781,7 @@ int TDBXML::LoadTableFile(PGLOBAL g, char *filename)
 
     // Initialize the implementation
     if (Docp->Initialize(g, Entry, Zipped)) {
-      sprintf(g->Message, MSG(INIT_FAILED), (Usedom) ? "DOM" : "libxml2");
+      snprintf(g->Message, sizeof(g->Message), MSG(INIT_FAILED), (Usedom) ? "DOM" : "libxml2");
       return RC_FX;
       } // endif init
 
@@ -864,7 +871,7 @@ bool TDBXML::Initialize(PGLOBAL g)
 			// Evaluate table xpath
 			if ((TabNode = Root->SelectSingleNode(g, tabpath))) {
 				if (TabNode->GetType() != XML_ELEMENT_NODE) {
-					sprintf(g->Message, MSG(BAD_NODE_TYPE), TabNode->GetType());
+					snprintf(g->Message, sizeof(g->Message), MSG(BAD_NODE_TYPE), TabNode->GetType());
 					goto error;
 				} // endif Type
 
@@ -881,12 +888,12 @@ bool TDBXML::Initialize(PGLOBAL g)
 				if (!(DBnode = Root->SelectSingleNode(g, tabpath))) {
 					// DB node does not exist yet; we cannot create it
 					// because we don't know where it should be placed
-					sprintf(g->Message, MSG(MISSING_NODE), XmlDB, Xfile);
+					snprintf(g->Message, sizeof(g->Message), MSG(MISSING_NODE), XmlDB, Xfile);
 					goto error;
 				} // endif DBnode
 
 				if (!(TabNode = DBnode->AddChildNode(g, Tabname))) {
-					sprintf(g->Message, MSG(FAIL_ADD_NODE), Tabname);
+					snprintf(g->Message, sizeof(g->Message), MSG(FAIL_ADD_NODE), Tabname);
 					goto error;
 				} // endif TabNode
 
@@ -931,7 +938,7 @@ bool TDBXML::Initialize(PGLOBAL g)
 					goto error;
 
 			} else {
-				sprintf(g->Message, MSG(FILE_UNFOUND), Xfile);
+				snprintf(g->Message, sizeof(g->Message), MSG(FILE_UNFOUND), Xfile);
 
 				if (Mode == MODE_READ) {
 					PushWarning(g, this);
@@ -943,7 +950,7 @@ bool TDBXML::Initialize(PGLOBAL g)
 
 		} else if (rc == RC_INFO) {
 			// Loading failed
-			sprintf(g->Message, MSG(LOADING_FAILED), Xfile);
+			snprintf(g->Message, sizeof(g->Message), MSG(LOADING_FAILED), Xfile);
 			goto error;
 		} else // (rc == RC_FX)
 			goto error;
@@ -967,7 +974,7 @@ bool TDBXML::Initialize(PGLOBAL g)
 
 
 		Docp->SetNofree(true);       // For libxml2
-#if defined(__WIN__)
+#if defined(_WIN32)
 	} catch (_com_error e) {
     // We come here if a DOM command threw an error
     char   buf[128];
@@ -976,12 +983,12 @@ bool TDBXML::Initialize(PGLOBAL g)
                              buf, sizeof(buf), NULL, NULL);
 
     if (rc)
-      sprintf(g->Message, "%s: %s", MSG(COM_ERROR), buf);
+      snprintf(g->Message, sizeof(g->Message), "%s: %s", MSG(COM_ERROR), buf);
     else
-      sprintf(g->Message, "%s hr=%x", MSG(COM_ERROR), e.Error());
+      snprintf(g->Message, sizeof(g->Message), "%s hr=%x", MSG(COM_ERROR), e.Error());
 
     goto error;
-#endif   // __WIN__
+#endif   // _WIN32
 #if !defined(UNIX)
   } catch(...) {
     // Other errors
@@ -1135,7 +1142,7 @@ int TDBXML::RowNumber(PGLOBAL g, bool b)
     /*******************************************************************/
     /*  Don't know how to retrieve RowID for expanded XML tables.      */
     /*******************************************************************/
-    sprintf(g->Message, MSG(NO_ROWID_FOR_AM),
+    snprintf(g->Message, sizeof(g->Message), MSG(NO_ROWID_FOR_AM),
                         GetAmName(g, GetAmType()));
     return 0;          // Means error
   } else
@@ -1257,7 +1264,7 @@ int TDBXML::ReadDB(PGLOBAL g)
     // Get the new row node
 		if (Nlist) {
 			if ((RowNode = Nlist->GetItem(g, Irow, RowNode)) == NULL) {
-				sprintf(g->Message, MSG(MISSING_ROWNODE), Irow);
+				snprintf(g->Message, sizeof(g->Message), MSG(MISSING_ROWNODE), Irow);
 				return RC_FX;
 			} // endif RowNode
 
@@ -1324,7 +1331,7 @@ int TDBXML::DeleteDB(PGLOBAL g, int irc)
     // Delete all rows
     for (Irow = 0; Irow < Nrow; Irow++)
       if ((RowNode = Nlist->GetItem(g, Irow, RowNode)) == NULL) {
-        sprintf(g->Message, MSG(MISSING_ROWNODE), Irow);
+        snprintf(g->Message, sizeof(g->Message), MSG(MISSING_ROWNODE), Irow);
         return RC_FX;
       } else {
         TabNode->DeleteChild(g, RowNode);
@@ -1510,13 +1517,17 @@ bool XMLCOL::ParseXpath(PGLOBAL g, bool mode)
   if (!mode)
     // Take care of an eventual extra column node a la html
     if (Tdbp->Colname) {
-      sprintf(pbuf, Tdbp->Colname, Rank + ((Tdbp->Usedom) ? 0 : 1));
-      strcat(pbuf, "/");
-      } // endif Colname
+      char *p = strstr(Tdbp->Colname, "%d");
+      if (p)
+        snprintf(pbuf, len + 3, "%.*s%d%s/", (int) (p - Tdbp->Colname), Tdbp->Colname,
+            Rank + (Tdbp->Usedom ? 0 : 1), p + 2);
+      else
+        snprintf(pbuf, len + 3, "%s/", Tdbp->Colname);
+    } // endif Colname
 
   if (Xname) {
     if (Type == 2) {
-      sprintf(g->Message, MSG(BAD_COL_XPATH), Name, Tdbp->Name);
+      snprintf(g->Message, sizeof(g->Message), MSG(BAD_COL_XPATH), Name, Tdbp->Name);
       return true;
     } else
       strcat(pbuf, Xname);
@@ -1548,7 +1559,7 @@ bool XMLCOL::ParseXpath(PGLOBAL g, bool mode)
       if (mode) {
         // For Update or Insert the Xpath must be explicit
         if (strchr("@/.*", *p)) {
-          sprintf(g->Message, MSG(XPATH_NOT_SUPP), Name);
+          snprintf(g->Message, sizeof(g->Message), MSG(XPATH_NOT_SUPP), Name);
           return true;
         } else
           Nodes[i] = p;
@@ -1559,7 +1570,7 @@ bool XMLCOL::ParseXpath(PGLOBAL g, bool mode)
       } // endfor i, p
 
     if (*p == '/' || *p == '.') {
-      sprintf(g->Message, MSG(XPATH_NOT_SUPP), Name);
+      snprintf(g->Message, sizeof(g->Message), MSG(XPATH_NOT_SUPP), Name);
       return true;
     } else if (*p == '@') {
       p++;                           // Remove the @ if mode
@@ -1620,7 +1631,7 @@ bool XMLCOL::ParseXpath(PGLOBAL g, bool mode)
 bool XMLCOL::SetBuffer(PGLOBAL g, PVAL value, bool ok, bool check)
   {
   if (!(To_Val = value)) {
-    sprintf(g->Message, MSG(VALUE_ERROR), Name);
+    snprintf(g->Message, sizeof(g->Message), MSG(VALUE_ERROR), Name);
     return true;
   } else if (Buf_Type == value->GetType()) {
     // Values are of the (good) column type
@@ -1639,7 +1650,7 @@ bool XMLCOL::SetBuffer(PGLOBAL g, PVAL value, bool ok, bool check)
   } else {
     // Values are not of the (good) column type
     if (check) {
-      sprintf(g->Message, MSG(TYPE_VALUE_ERR), Name,
+      snprintf(g->Message, sizeof(g->Message), MSG(TYPE_VALUE_ERR), Name,
               GetTypeName(Buf_Type), GetTypeName(value->GetType()));
       return true;
       } // endif check
@@ -1692,7 +1703,7 @@ void XMLCOL::ReadColumn(PGLOBAL g)
   if (ValNode) {
     if (ValNode->GetType() != XML_ELEMENT_NODE &&
         ValNode->GetType() != XML_ATTRIBUTE_NODE) {
-      sprintf(g->Message, MSG(BAD_VALNODE), ValNode->GetType(), Name);
+      snprintf(g->Message, sizeof(g->Message), MSG(BAD_VALNODE), ValNode->GetType(), Name);
 			throw (int)TYPE_AM_XML;
 		} // endif type
 
@@ -1802,6 +1813,9 @@ void XMLCOL::WriteColumn(PGLOBAL g)
     else if (Tdbp->Clist)
       ColNode = NULL;
 
+    // refresh CList in case its Listp was freed in SelectSingleNode above
+    if (Tdbp->Clist)
+      Tdbp->RowNode->SelectNodes(g, Tdbp->Colname, Tdbp->Clist);
     } // endfor i
 
   /*********************************************************************/
@@ -1858,7 +1872,7 @@ void XMLCOL::WriteColumn(PGLOBAL g)
   p = Value->GetCharString(buf);
 
   if (strlen(p) > (unsigned)Long) {
-    sprintf(g->Message, MSG(VALUE_TOO_LONG), p, Name, Long);
+    snprintf(g->Message, sizeof(g->Message), MSG(VALUE_TOO_LONG), p, Name, Long);
 		throw (int)TYPE_AM_XML;
 	} else
     strcpy(Valbuf, p);
@@ -1899,7 +1913,7 @@ void XMULCOL::ReadColumn(PGLOBAL g)
 
       if (N > Tdbp->Limit) {
         N = Tdbp->Limit;
-        sprintf(g->Message, "Multiple values limited to %d", Tdbp->Limit);
+        snprintf(g->Message, sizeof(g->Message), "Multiple values limited to %d", Tdbp->Limit);
         PushWarning(g, Tdbp);
         } // endif N
 
@@ -1908,7 +1922,7 @@ void XMULCOL::ReadColumn(PGLOBAL g)
 
         if (ValNode->GetType() != XML_ELEMENT_NODE &&
             ValNode->GetType() != XML_ATTRIBUTE_NODE) {
-          sprintf(g->Message, MSG(BAD_VALNODE), ValNode->GetType(), Name);
+          snprintf(g->Message, sizeof(g->Message), MSG(BAD_VALNODE), ValNode->GetType(), Name);
 					throw (int)TYPE_AM_XML;
 				} // endif type
 
@@ -2044,7 +2058,7 @@ void XMULCOL::WriteColumn(PGLOBAL g)
         len = Nlx->GetLength();
 
         if (len > 1 && !Tdbp->Xpand) {
-          sprintf(g->Message, MSG(BAD_VAL_UPDATE), Name);
+          snprintf(g->Message, sizeof(g->Message), MSG(BAD_VAL_UPDATE), Name);
 					throw (int)TYPE_AM_XML;
 				} else
           ValNode = Nlx->GetItem(g, Tdbp->Nsub, Vxnp);
@@ -2119,7 +2133,7 @@ void XMULCOL::WriteColumn(PGLOBAL g)
   p = Value->GetCharString(buf);
 
   if (strlen(p) > (unsigned)Long) {
-    sprintf(g->Message, MSG(VALUE_TOO_LONG), p, Name, Long);
+    snprintf(g->Message, sizeof(g->Message), MSG(VALUE_TOO_LONG), p, Name, Long);
 		throw (int)TYPE_AM_XML;
 	} else
     strcpy(Valbuf, p);
@@ -2246,7 +2260,7 @@ void XPOSCOL::WriteColumn(PGLOBAL g)
   p = Value->GetCharString(buf);
 
   if (strlen(p) > (unsigned)Long) {
-    sprintf(g->Message, MSG(VALUE_TOO_LONG), p, Name, Long);
+    snprintf(g->Message, sizeof(g->Message), MSG(VALUE_TOO_LONG), p, Name, Long);
 		throw (int)TYPE_AM_XML;
 	} else
     strcpy(Valbuf, p);
