@@ -1,4 +1,4 @@
-/* Copyright (C) 2019, 2020 MariaDB Corporation Ab
+/* Copyright (C) 2019, 2021 MariaDB Corporation Ab
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -640,6 +640,7 @@ int ha_s3::open(const char *name, int mode, uint open_flags)
       file->dfile.big_block_size= file->s->kfile.big_block_size=
         file->s->bitmap.file.big_block_size= file->s->base.s3_block_size;
       file->s->kfile.head_blocks= file->s->base.keystart / file->s->block_size;
+      file->s->no_status_updates= in_alter_table == S3_NO_ALTER;
     }
   }
   open_args= 0;
@@ -887,10 +888,11 @@ int ha_s3::discover_check_version()
   Update the .frm file in S3
 */
 
-static int s3_notify_tabledef_changed(handlerton *hton __attribute__((unused)),
+static int s3_notify_tabledef_changed(handlerton *,
                                       LEX_CSTRING *db, LEX_CSTRING *table,
                                       LEX_CUSTRING *frm,
-                                      LEX_CUSTRING *org_tabledef_version)
+                                      LEX_CUSTRING *org_tabledef_version,
+                                      handler *)
 {
   char aws_path[AWS_PATH_LENGTH];
   S3_INFO s3_info;
@@ -898,6 +900,9 @@ static int s3_notify_tabledef_changed(handlerton *hton __attribute__((unused)),
   int error= 0;
   DBUG_ENTER("s3_notify_tabledef_changed");
 
+  if (strstr(table->str, "#P#"))
+    DBUG_RETURN(0);                             // Ignore partitions
+
   if (s3_info_init(&s3_info))
     DBUG_RETURN(0);
   if (!(s3_client= s3_open_connection(&s3_info)))
@@ -916,7 +921,7 @@ static int s3_notify_tabledef_changed(handlerton *hton __attribute__((unused)),
            NullS);
 
   if (s3_put_object(s3_client, s3_info.bucket.str, aws_path, (uchar*) frm->str,
-                         frm->length, 0))
+                    frm->length, 0))
     error= 2;
 
 err:
@@ -1107,6 +1112,6 @@ maria_declare_plugin(s3)
   status_variables,             /* status variables */
   system_variables,             /* system variables */
   "1.0",                        /* string version   */
-  MariaDB_PLUGIN_MATURITY_GAMMA /* maturity         */
+  MariaDB_PLUGIN_MATURITY_STABLE/* maturity         */
 }
 maria_declare_plugin_end;
