@@ -1,6 +1,6 @@
 /*
    Copyright (c) 2000, 2018, Oracle and/or its affiliates.
-   Copyright (c) 2009, 2017, MariaDB Corporation.
+   Copyright (c) 2009, 2022, MariaDB Corporation.
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -33,6 +33,7 @@
 #include "sql_table.h"                          // tablename_to_filename
 #include "sql_class.h"                          // THD
 #include "debug_sync.h"
+#include "sql_debug.h"
 
 ulonglong myisam_recover_options;
 static ulong opt_myisam_block_size;
@@ -121,6 +122,28 @@ static void debug_wait_for_kill(const char *info)
   thd_proc_info(thd, prev_info);
   DBUG_VOID_RETURN;
 }
+
+
+class Debug_key_myisam: public Debug_key
+{
+public:
+  Debug_key_myisam() { }
+
+  static void print_keys_myisam(THD *thd, const char *where,
+                                const TABLE *table,
+                                const MI_KEYDEF *keydef, uint count)
+  {
+    for (uint i= 0; i < count; i++)
+    {
+      Debug_key_myisam tmp;
+      if (!tmp.append(where) &&
+          !tmp.append_key(table->s->key_info[i].name, keydef[i].flag))
+        tmp.print(thd);
+      print_keysegs(thd, keydef[i].seg, keydef[i].keysegs);
+    }
+  }
+};
+
 #endif
 
 /*****************************************************************************
@@ -707,7 +730,7 @@ static int compute_vcols(MI_INFO *info, uchar *record, int keynum)
   {
     Field *f= table->field[kp->fieldnr - 1];
     if (f->vcol_info && !f->vcol_info->stored_in_db)
-      table->update_virtual_field(f);
+      table->update_virtual_field(f, false);
   }
   mysql_mutex_unlock(&info->s->intern_lock);
   return 0;
@@ -837,6 +860,13 @@ int ha_myisam::open(const char *name, int mode, uint test_if_locked)
       /* purecov: end */
     }
   }
+
+  DBUG_EXECUTE_IF("key",
+    Debug_key_myisam::print_keys_myisam(table->in_use,
+                                        "ha_myisam::open: ",
+                                        table, file->s->keyinfo,
+                                        file->s->base.keys);
+  );
   
   if (test_if_locked & (HA_OPEN_IGNORE_IF_LOCKED | HA_OPEN_TMP_TABLE))
     (void) mi_extra(file, HA_EXTRA_NO_WAIT_LOCK, 0);
@@ -978,7 +1008,8 @@ void ha_myisam::setup_vcols_for_repair(HA_CHECK *param)
       return;
     file->s->vreclength= new_vreclength;
   }
-  DBUG_ASSERT(file->s->base.reclength < file->s->vreclength);
+  DBUG_ASSERT(file->s->base.reclength < file->s->vreclength ||
+              !table->s->stored_fields);
   param->fix_record= compute_vcols;
   table->use_all_columns();
 }
@@ -2233,6 +2264,15 @@ int ha_myisam::create(const char *name, TABLE *table_arg,
 
   if ((error= table2myisam(table_arg, &keydef, &recinfo, &record_count)))
     DBUG_RETURN(error); /* purecov: inspected */
+
+#ifndef DBUG_OFF
+  DBUG_EXECUTE_IF("key",
+    Debug_key_myisam::print_keys_myisam(table_arg->in_use,
+                                        "ha_myisam::create: ",
+                                        table_arg, keydef, share->keys);
+  );
+#endif
+
   bzero((char*) &create_info, sizeof(create_info));
   create_info.max_rows= share->max_rows;
   create_info.reloc_rows= share->min_rows;
