@@ -1,5 +1,5 @@
 /* Copyright (C) 2008-2020 Kentoku Shiba
-   Copyright (C) 2019-2020 MariaDB corp
+   Copyright (C) 2019, 2022, MariaDB Corporation.
 
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
@@ -101,6 +101,8 @@ typedef st_spider_result SPIDER_RESULT;
 
 #define SPIDER_SQL_DOT_STR "."
 #define SPIDER_SQL_DOT_LEN (sizeof(SPIDER_SQL_DOT_STR) - 1)
+#define SPIDER_SQL_PERCENT_STR "%"
+#define SPIDER_SQL_PERCENT_LEN (sizeof(SPIDER_SQL_PERCENT_STR) - 1)
 #define SPIDER_SQL_HYPHEN_STR "-"
 #define SPIDER_SQL_HYPHEN_LEN (sizeof(SPIDER_SQL_HYPHEN_STR) - 1)
 
@@ -172,6 +174,8 @@ typedef st_spider_result SPIDER_RESULT;
 #define SPIDER_SQL_IN_LEN (sizeof(SPIDER_SQL_IN_STR) - 1)
 #define SPIDER_SQL_NOT_IN_STR "not in("
 #define SPIDER_SQL_NOT_IN_LEN (sizeof(SPIDER_SQL_NOT_IN_STR) - 1)
+#define SPIDER_SQL_LIKE_STR " like "
+#define SPIDER_SQL_LIKE_LEN (sizeof(SPIDER_SQL_LIKE_STR) - 1)
 #define SPIDER_SQL_NOT_LIKE_STR "not like"
 #define SPIDER_SQL_NOT_LIKE_LEN (sizeof(SPIDER_SQL_NOT_LIKE_STR) - 1)
 #define SPIDER_SQL_AS_CHAR_STR " as char"
@@ -853,13 +857,10 @@ public:
   virtual int append_escaped_name_quote(
     spider_string *str
   ) = 0;
-  virtual int append_column_value(
-    ha_spider *spider,
-    spider_string *str,
-    Field *field,
-    const uchar *new_ptr,
-    CHARSET_INFO *access_charset
-  ) = 0;
+  virtual int append_column_value(ha_spider *spider, spider_string *str,
+                                  Field *field, const uchar *new_ptr,
+                                  bool is_like,
+                                  CHARSET_INFO *access_charset)= 0;
   virtual int append_trx_isolation(
     spider_string *str,
     int trx_isolation
@@ -1355,6 +1356,7 @@ public:
   SPIDER_LINK_IDX_CHAIN *link_idx_chain;
 #endif
   bool strict_group_by;
+  bool no_where_cond;
   spider_db_handler(ha_spider *spider, spider_db_share *db_share) :
     dbton_id(db_share->dbton_id), spider(spider), db_share(db_share),
     first_link_idx(-1) {}
